#include "option.h"
#include "tools.h"
#include "plugin.h"
#include "resource.h"

#undef min
#undef max

using namespace std;

static_assert( sizeof(INT32) == sizeof(float), "float size mismatch." );

namespace option
{
	namespace detail
	{
		Base::Base( HWND hwnd ): hwnd_(tools::GetTopParent(hwnd)), config_(tools::GetString(IDS_SYSNAME)) {}
	}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	//	f[^ǂݎ
	LONG	Reader::QueryValue( wchar_t const* name, DWORD type, void* data, DWORD& size ) const
	{
		return	Editor_RegQueryValue( hwnd_, EEREG_EMEDITORPLUGIN, config_.c_str(), name, type, reinterpret_cast<BYTE*>(data), &size, 0 );
	}

	//	32bitǂݎ
	INT32	Reader::Int32( wchar_t const* name, INT32 defaultValue ) const
	{
		DWORD	value( defaultValue  );
		DWORD	size(  sizeof(value) );
		return	(QueryValue( name, REG_DWORD, &value, size ) == ERROR_SUCCESS && size == sizeof(value)) ? value : defaultValue;
	}

	//	_lǂݎ
	bool	Reader::Bool( wchar_t const* name, bool defaultValue ) const
	{
		return	Int32( name, defaultValue ? 1 : 0 ) != 0;
	}

	//	Px_ǂݎ
	float	Reader::Float( wchar_t const* name, float defaultValue ) const
	{
		INT32 value( Int32( name, *reinterpret_cast<INT32*>(&defaultValue) ) );
		return	*reinterpret_cast<float*>(&value);
	}

	//	ǂݎ
	wstring	Reader::String( wchar_t const* name, wchar_t const*	defaultValue ) const
	{
		wchar_t	buf[4096];
		DWORD	size( sizeof(buf) - sizeof(wchar_t) );
		if( QueryValue( name, REG_SZ, buf, size ) == ERROR_SUCCESS ){
			buf[std::extent<decltype(buf)>::value - 1] = L'\0';
			return	buf;
		}
		return	defaultValue ? defaultValue : L"";
	}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	//	f[^
	LONG	Writer::SetValue( wchar_t const* name, DWORD type, const void* data, DWORD size )
	{
		return	Editor_RegSetValue(	hwnd_, EEREG_EMEDITORPLUGIN, config_.c_str(), name, type, reinterpret_cast<BYTE const*>(data), size, 0 );
	}

	//	32bit
	bool	Writer::Int32( wchar_t const* name, INT32 value )
	{
		return	SetValue( name, REG_DWORD, &value, sizeof(UINT32) ) != ERROR_SUCCESS;
	}

	//	_l
	bool	Writer::Bool( wchar_t const* name, bool value )
	{
		return	Int32( name, value ? 1 : 0 );
	}

	//	Px_
	bool	Writer::Float( wchar_t const* name, float value )
	{
		return	Int32( name, *reinterpret_cast<INT32*>(&value) );
	}

	//	񏑂
	bool	Writer::String( wchar_t const* name, wchar_t const* value )
	{
		return	SetValue( name, REG_SZ, value ? value : L"", static_cast<DWORD>((value ? (wcslen(value) + 1) : 1) * sizeof(wchar_t)) ) != ERROR_SUCCESS;
	}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////
}
